@@ -32,7 +32,7 @@
 	     (k start (+ k 1)))
 	    ((= k len) dest)
 	  (set! (dest i) (src k)))
-	(do ((i (+ at (- len start 1)) (- i 1))
+	(do ((i (- (+ at len) start 1) (- i 1))
 	     (k (- len 1) (- k 1)))
 	    ((< k start) dest)
 	  (set! (dest i) (src k))))))
@@ -45,10 +45,10 @@
 	  (apply #_make-hash-table args))))
 
 (define bytevector byte-vector)
-(define ->bytevector ->byte-vector)
 (define bytevector? byte-vector?)
 (define make-bytevector make-byte-vector)
-
+(define bytevector-ref byte-vector-ref)
+(define bytevector-set! byte-vector-set!)
 (define bytevector-copy! vector-copy!)
 (define string-copy! vector-copy!)
 (define (bytevector->list bv) (map values bv))
@@ -74,10 +74,7 @@
 (define (digit-value c) (and (char-numeric? c) (- (char->integer c) (char->integer #\0))))
 
 
-(define +inf.0 inf.0)
-(define +nan.0 nan.0)
 (define (finite? n) (and (number? n) (not (nan? n)) (not (infinite? n))))
-
 (define exact-integer? integer?)	
 (define (exact-integer-sqrt i) (let ((sq (floor (sqrt i)))) (values sq (- i (* sq sq)))))
 (define inexact exact->inexact)
@@ -94,19 +91,19 @@
 (define (port? p) (or (input-port? p) (output-port? p)))
 (define binary-port? port?)
 (define textual-port? port?)
-(define (close-port p) (if (input-port? p) (close-input-port p) (close-output-port p)))
+(define (close-port p) ((if (input-port? p) close-input-port close-output-port) p))
 (define open-binary-input-file open-input-file)
 (define open-binary-output-file open-output-file)
-(define (call-with-port port proc) 
-  ((if (input-port? port) call-with-input-file call-with-output-file) port proc))
+(define (call-with-port port proc) ((if (input-port? port) call-with-input-file call-with-output-file) port proc))
 
 
 (define (bytevector-u8-ref b k) (b k))
 (define (bytevector-u8-set! b k c) (set! (b k) c))
+(define bytevector-u8 (dilambda (lambda (b k) (b k)) (lambda (b k c) (set! (b k) c))))
 (define bytevector-length length)
-(define (bytevector-copy . args) (->byte-vector (apply string-copy args)))
-(define (bytevector-append . args) (->byte-vector (apply string-append args)))
-(define write-bytevector write-string)
+(define bytevector-copy vector-copy!)
+(define bytevector-append append)
+(define write-bytevector write)
 (define* (read-bytevector! bv port (start 0) end)
   (let ((lim (or end (length bv)))
 	(pt (or port (current-input-port))))
@@ -117,8 +114,8 @@
 	 bv)
       (set! (bv i) c))))
 (define* (read-bytevector k port)
-  (read-bytevector! (->byte-vector (make-string k)) port))
-(define (get-output-bytevector port) (->byte-vector (get-output-string port)))
+  (read-bytevector! (make-byte-vector k) port))
+(define (get-output-bytevector port) (string->byte-vector (get-output-string port)))
 (define open-input-bytevector open-input-string)
 (define open-output-bytevector open-output-string)
 (define read-u8 read-byte)
@@ -126,11 +123,11 @@
 (define u8-ready? char-ready?) 
 (define peek-u8 peek-char)
 (define* (utf8->string v (start 0) end) (substring v start (or end (length v))))
-(define* (string->utf8 s (start 0) end) (->byte-vector (utf8->string s start end)))
+(define* (string->utf8 s (start 0) end) (string->byte-vector (utf8->string s start end)))
 (define write-simple write)
 
 (define (eof-object) #<eof>)
-(define (features) *features*)
+(define-macro (features) '*features*) ; needs to be the local *features*
 
 
 (define (with-exception-handler handler thunk) (catch #t thunk handler))
@@ -148,6 +145,7 @@
 
 
 (define interaction-environment curlet)
+;; for null-environment see stuff.scm
 (define-macro (include . files) 
   `(begin
      ,@(map (lambda (file)
@@ -159,16 +157,56 @@
 ;;   or (format #f "~^ this is a comment ")
 
 
-(define-macro (define-values vars . body)
-  `(apply begin (map (lambda (var val) `(define ,var ,val)) ',vars (list (begin ,@body)))))
+(define-macro (define-values vars expression)
+  `(if (not (null? ',vars))
+       (varlet (curlet) ((lambda ,vars (curlet)) ,expression))))
 
-(define-macro (let*-values vars . body)
-  `(let () 
-     ,@(map (lambda (nvars . nbody)
-              `(apply define-values ',nvars ',@nbody))
-            (map car vars) 
-            (map cdr vars))
-     ,@body))
+#|
+(define-macro (define*-values vars expression) ; same but allows defaults for the vars
+  `(if (not (null? ',vars))
+       (varlet (curlet) ((lambda* ,vars (curlet)) ,expression))))
+
+(define-macro (define-values vars . body) ; but the spec says "<expression>" here
+  `(apply begin (map (lambda (var val) `(define ,var ,val)) ',vars (list (begin ,@body)))))
+|#
+
+(define-macro (let-values vars . body)
+  (if (and (pair? vars)
+	   (pair? (car vars))
+	   (null? (cdar vars)))
+      `((lambda ,(caar vars)
+	  ,@body)
+	,(cadar vars))
+      `(with-let (apply sublet (curlet)
+			(list ,@(map (lambda (v)
+				       `((lambda ,(car v)
+					   (values ,@(map (lambda (name)
+							    (values (symbol->keyword name) name))
+							  (let args->proper-list ((args (car v)))
+							    (cond ((symbol? args)	(list args))
+								  ((not (pair? args))	args)
+								  ((pair? (car args))	(cons (caar args) (args->proper-list (cdr args))))
+								  (else                 (cons (car args) (args->proper-list (cdr args)))))))))
+					 ,(cadr v)))
+				     vars)))
+	 ,@body)))
+
+(define-macro (let*-values vals . body)
+  (let ((args ())
+	(exprs ()))
+    (for-each
+     (lambda (arg+expr)
+       (set! args (cons (car arg+expr) args))
+       (set! exprs (cons (cadr arg+expr) exprs)))
+     vals)
+    (let ((form `((lambda ,(car args) ,@body) ,(car exprs))))
+      (if (not (null? (cdr args)))
+	  (for-each
+	   (lambda (arg expr)
+	     (set! form (list (list 'lambda arg form) expr)))
+	   (cdr args)
+	   (cdr exprs)))
+      form)))
 
 
 ;; case-lambda       
@@ -299,7 +337,7 @@
 (define-macro (delay-force expr) 
   `(make-promise #f (lambda () ,expr)))
 (define-macro (r7rs-delay expr) ; "delay" is taken damn it
-  `(delay-force (make-promise #t (lambda () ,expr))))
+  (list 'delay-force (list 'make-promise #t (list 'lambda () expr))))
 (define (make-promise done? proc) 
   (list (cons done? proc)))
 (define (force promise)
@@ -342,7 +380,11 @@
     (cons 'current-second g_time)))
 
 
-(define get-environment-variable (*libc* 'getenv))
+(define (get-environment-variable x)
+  (let ((val ((*libc* 'getenv) x)))
+    (and (string? val)
+	 (> (length val) 0)
+	 val)))
 (define get-environment-variables (*libc* 'getenvs))
 (define (r7rs-file-exists? arg) (= ((*libc* 'access) arg (*libc* 'F_OK)) 0))
 (define r7rs-delete-file (*libc* 'unlink))
@@ -351,7 +393,7 @@
 (define (cpu-architecture) (cadr ((*libc* 'uname))))
 (define (machine-name) (caddr ((*libc* 'uname))))
 (define (os-version) (string-append (list-ref ((*libc* 'uname)) 3) " " (list-ref ((*libc* 'uname)) 4)))
-(define (implementation-name) "s7")
+(define (implementation-name) (copy "s7"))
 (define (implementation-version) (substring (s7-version) 3 7))
 
 ;; command-line is problematic: s7 has no access to the caller's "main" function, and
@@ -375,54 +417,55 @@
 
 ;; other minor differences: 
 ;;  in s7, single-quote can occur in a name
-;;  s7 doesn't currently implement #\xxxx characters
 
 
 ;; records
 (define-macro (define-record-type type make ? . fields)
   (let ((new-type (if (pair? type) (car type) type))
-	(inherited (if (pair? type) (cdr type) ())))
+	(inherited (if (pair? type) (cdr type) ()))
+	(obj (gensym))
+	(new-obj (gensym)))
     `(begin
        (define-class ,new-type ,inherited   ; from stuff.scm
          (map (lambda (f) (if (pair? f) (car f) f)) ',fields))
        
-       (define (,? obj)    ; perhaps the define-class type predicate should use this 
-         (define (search-inherited obj type)
+       (define (,? ,obj)    ; perhaps the define-class type predicate should use this 
+         (define (search-inherited ,obj type)
 	   (define (search-inheritors objs type)
 	     (and (pair? objs)
 		  (or (search-inherited (car objs) type)
 		      (search-inheritors (cdr objs) type))))
-	   (or (eq? (obj 'class-name) type)
-	       (search-inheritors (obj 'inherited) type)))
-         (and (let? obj)
-	      (search-inherited obj ',new-type)))
+	   (or (eq? (,obj 'class-name) type)
+	       (search-inheritors (,obj 'inherited) type)))
+         (and (let? ,obj)
+	      (search-inherited ,obj ',new-type)))
        
        (define ,make 
-         (let ((new-obj (copy ,new-type)))
+         (let ((,new-obj (copy ,new-type)))
 	   ,@(map (lambda (slot)
-		    `(set! (new-obj ',slot) ,slot))
+		    `(set! (,new-obj ',slot) ,slot))
 		  (cdr make))
-	   new-obj))
+	   ,new-obj))
        
        ,@(map
 	  (lambda (field)
-	    (if (pair? field)
-	        (if (null? (cdr field))
-		    (values)
-		    (if (null? (cddr field))
-		        `(define (,(cadr field) obj)
-                           (if (not (,? obj)) 
-                               (error 'wrong-type-arg "~S should be a ~A" obj ',type))
-                           (obj ',(car field)))
-		        `(begin
-			   (define (,(cadr field) obj)
-			     (if (not (,? obj)) 
-				 (error 'wrong-type-arg "~S should be a ~A" obj ',type))
-                             (obj ',(car field)))
-			   (define (,(caddr field) obj val)
-			     (if (not (,? obj)) 
-				 (error 'wrong-type-arg "~S should be a ~A" obj ',type))
-                             (set! (obj ',(car field)) val)))))))
+	    (when (pair? field)
+	      (if (null? (cdr field))
+		  (values)
+		  (if (null? (cddr field))
+		      `(define (,(cadr field) ,obj)
+			 (if (not (,? ,obj)) 
+			     (error 'wrong-type-arg "~S should be a ~A" ,obj ',type))
+			 (,obj ',(car field)))
+		      `(begin
+			 (define (,(cadr field) ,obj)
+			   (if (not (,? ,obj)) 
+			       (error 'wrong-type-arg "~S should be a ~A" ,obj ',type))
+			   (,obj ',(car field)))
+			 (define (,(caddr field) ,obj val)
+			   (if (not (,? ,obj)) 
+			       (error 'wrong-type-arg "~S should be a ~A" ,obj ',type))
+			   (set! (,obj ',(car field)) val)))))))
 	  fields)
        ',new-type)))
 
